from ...clients.sqs_client.post_message import SyncSQSPusher
from ...clients.sqs_client.polling import SQS_Manager
from ...clients.sqs_client.initiator import SQSInitiator
from ...clients.sqs_client.queue_config import QUEUE_CONFIG, QUEUE_NAMES



from redis import Redis
from typing import Any, List, Optional, Generator, cast
import json
from ...interfaces.interfaces_pd import SSEPayload_PM
from ...ioc.singleton import SingletonMeta


class SyncRedisClient(metaclass=SingletonMeta):
    def __init__(
        self,
        redis_connection: Optional[Redis] = None,
    ):
        if not hasattr(self, "redis") and isinstance(redis_connection, Redis):
            self.redis = redis_connection
        self.pubsub = self.redis.pubsub()

    def set(self, key: str, value: Any, expiry: Optional[int] = None) -> bool:
        return bool(self.redis.set(key, value, ex=expiry))

    def get(self, key: str) -> Optional[bytes]:
        return cast(Optional[bytes], self.redis.get(key))

    def delete(self, key: str) -> int:
        return cast(int, self.redis.delete(key))

    def delete_all_keys(self, keys: List[str]) -> int:
        return cast(int, self.redis.delete(*keys))

    def get_keys(self, pattern: str = "*") -> List[str]:
        return cast(List[str], self.redis.keys(pattern))

    def send_command(self, *commands: str) -> Any:
        return self.redis.execute_command(*commands)

    def publish(self, channel: str, message: SSEPayload_PM) -> int:
        return cast(
            int,
            self.redis.publish(
                channel=channel, message=json.dumps(message.model_dump())
            ),
        )

    def subscribe(self, channel: str) -> None:
        self.pubsub.subscribe(channel)

    def unsubscribe(self, channel: str) -> None:
        self.pubsub.unsubscribe(channel)

    def get_message(self, timeout: float = 1.0) -> Optional[dict]:
        return self.pubsub.get_message(timeout=timeout)

    def listen(self, channel: str) -> Generator[dict, None, None]:
        self.subscribe(channel=channel)
        try:
            while True:
                message = self.get_message(timeout=1.0)
                if message is not None:
                    yield message
                # Using time.sleep instead of asyncio.sleep for synchronous operation
                import time

                time.sleep(0.1)
        finally:
            self.unsubscribe(channel=channel)

    def close_connection(self):
        self.pubsub.close()
        self.redis.close()



from typing import List, Optional
from .base_logger import APP_LOGGER


class LogExecution:
    def __init__(self, vars_to_log: Optional[List[str]] = None) -> None:
        self.vars_to_log = vars_to_log

    def __call__(self, func):
        def wrapper(instance, *args, **kwargs):
            class_name = instance.__class__.__name__
            function_name = func.__name__
            APP_LOGGER.info(f"Executing {function_name} method of class {class_name}.")
            try:
                result = func(instance, *args, **kwargs)
                instance_variables = (
                    {
                        var: getattr(instance, var)
                        for var in self.vars_to_log
                        if hasattr(instance, var)
                    }
                    if self.vars_to_log
                    else None
                )
                if result or instance_variables:
                    APP_LOGGER.info(
                        {
                            "instance_variables": instance_variables,
                            "result": result,
                        }
                    )
                return result
            except Exception:
                raise

        return wrapper



from typing import Any, Dict
import uuid
from pydantic import BaseModel, Field


class Headers_PM(BaseModel):
    correlationid: str = Field(default_factory=lambda: str(uuid.uuid4()))
    username: str = "not_applicable"
    authorization: str = Field(default="")

    def model_dump(self, exclude_fields={}, **kwargs) -> Dict[str, Any]:
        return super().model_dump(**kwargs, exclude=exclude_fields)


class RequestContext_PM(BaseModel):
    request_trace_id: str
    url_path: str
    method: str


class SSEPayload_PM(BaseModel):
    event: str
    data: Any



import json
import asyncio
from functools import partial
from typing import Dict, final, List, Union, Callable, Awaitable
from types import MethodType

from ...logging.base_logger import APP_LOGGER
from ...interfaces.interfaces_th import SQSClientCallBackResponse_TH
from ...ioc.singleton import Para_SingletonMeta


@final
class MessageHandler(metaclass=Para_SingletonMeta):
    QUEUE_TO_MESSAGE_HANDLER_MAP: Dict[str, Dict[str, Callable[[], Awaitable[str]]]]
    QUEUE_CONFIG: Dict[str, Dict[str, bool]]

    def __init__(self, queue_name: str) -> None:
        self.message_handler_map = self.QUEUE_TO_MESSAGE_HANDLER_MAP[queue_name]
        if self.message_handler_map is None:
            raise ValueError(f"No Message Handler for: {queue_name}")

    def task_callback(self, task, message_entry):
        if task.done() and task.exception():
            APP_LOGGER.error(
                f"'SQS Action Handler' Callback Response : {task.exception()} || {message_entry} "
            )

        else:
            callback_response: SQSClientCallBackResponse_TH = json.loads(task.result())
            allSuccess = callback_response["allSuccess"]
            if allSuccess:
                self.entries_to_delete.append(message_entry)
            APP_LOGGER.info(
                f"'SQS Action Handler' CallBack Response: {callback_response} || {message_entry}"
            )

    async def handle_messages_in_batch(
        self, messages_batch_payload: List[Dict[str, str]]
    ):
        self.entries_to_delete = []
        awaitables = []
        for message in messages_batch_payload:
            try:
                message_body = json.loads(message["Body"])
                message_context: Dict[str, str] = message_body["context"]
                message_metadata: Dict[str, Union[dict, str]] = dict(
                    message_entry=dict(
                        Id=message["MessageId"],
                        ReceiptHandle=message["ReceiptHandle"],
                        correlationid=message_context["correlationid"],
                    ),
                    action=message_body["action"],
                    payload=message_body["payload"],
                )
                action: str = str(message_metadata.get("action"))

                message_handler = self.message_handler_map.get(action)

                # NOTE : Enable to verify message_context of all messages if they are comming from valid sources .ignore_expiry=False if you want to validate jwt token.
                # try:
                #     await asyncio.to_thread(
                #         lambda payload, model: model(**payload, ignore_expiry=True),
                #         message_context,
                #         HeaderModel,
                #     )
                # except ValidationError as e:
                #     APP_LOGGER.error(f"{message_metadata} || Error: {e}")
                #     self.entries_to_delete.append(message_metadata.get("message_entry"))
                #     continue

                APP_LOGGER.info(message_metadata)

                if isinstance(message_handler, MethodType):
                    task = asyncio.create_task(
                        message_handler(
                            payload=message_metadata["payload"], headers=message_context
                        )
                    )
                    callback_with_args = partial(
                        self.task_callback,
                        message_entry=message_metadata["message_entry"],
                    )
                    task.add_done_callback(callback_with_args)
                    awaitables.append(task)
                else:
                    APP_LOGGER.warning(
                        f"{message_metadata} || No 'SQS Action Handler' found ."
                    )
            except Exception as e:
                APP_LOGGER.error(f"Error processing message: {e}")
                # self.entries_to_delete.append(message_metadata.get("message_entry"))
                continue

        if awaitables:
            await asyncio.gather(*awaitables, return_exceptions=True)

        return self.entries_to_delete



from ast import List
from functools import lru_cache
import json
from fastapi import HTTPException
from jwcrypto import jwk
import jwt
from decouple import config
from typing import Dict, List, Optional

from ..interfaces.interfaces_th import Jwk_TH
from ..logging.base_logger import APP_LOGGER


class JwtValdationUtils:
    JWT_ALGORITHM = "RS256" if config("AUTH_TOKEN_ALGORITHM") == "RS256" else "HS256"

    @staticmethod
    def is_token_expired(token: str) -> bool:
        try:
            JwtValdationUtils.validate_token(token, verify_exp=True)
            return False
        except HTTPException as e:
            if e.status_code == 401 and e.detail == "Token has expired":
                return True
            raise e

    @lru_cache(maxsize=1)
    @staticmethod
    def __load_jwks() -> List[Jwk_TH]:
        return json.loads(str(config("JWKS")))

    @classmethod
    def _get_public_key(cls, kid: str) -> Optional[str]:
        public_key = next((key for key in cls.__load_jwks() if key["kid"] == kid), None)
        if public_key:
            return jwk.JWK(**public_key).export_to_pem().decode("utf-8")
        return None

    @classmethod
    def validate_token(
        cls,
        token: str,
        verify_exp: bool = False,
        verify_aud: bool = False,
    ) -> Dict:
        try:
            default_options = dict()
            if not verify_exp:
                default_options.update(verify_exp=False)
            if not verify_aud:
                default_options.update(verify_aud=False)

            unverified_header = jwt.get_unverified_header(token)
            public_key = cls._get_public_key(unverified_header["kid"])

            if not public_key:
                raise ValueError("No matching public key found")

            return jwt.decode(
                token,
                public_key,
                algorithms=[cls.JWT_ALGORITHM],
                options=default_options,
            )
        except jwt.ExpiredSignatureError:
            raise HTTPException(status_code=401, detail="Token has expired")
        except (jwt.PyJWTError, ValueError) as e:
            APP_LOGGER.error(f"Token validation failed: {str(e)}")
            raise HTTPException(status_code=401, detail="Invalid token")



from enum import StrEnum


class FoundationModels(StrEnum):
    LLAMA_3_8B_INSTRUCT = "meta.llama3-8b-instruct-v1:0"
    LLAMA_3_1_8B_INSTRUCT = "meta.llama3-1-8b-instruct-v1:0"
    LLAMA_3_70B_INSTRUCT = "meta.llama3-70b-instruct-v1:0"
    LLAMA_3_1_70B_INSTRUCT = "meta.llama3-1-70b-instruct-v1:0"



from asyncio import CancelledError
from typing import FrozenSet
import uuid
import json
import sys
import traceback
from decouple import config
from fastapi import HTTPException, Request
from fastapi.responses import JSONResponse
from decouple import config
from starlette.types import ASGIApp, Scope, Receive, Send

# modules

from ..interfaces.interfaces_pd import Headers_PM, RequestContext_PM
from ..utils.jwt_validation import JwtValdationUtils
from ..logging.base_logger import APP_LOGGER
from ..context.vars import (
    headers_context,
    request_context,
    request_params_context,
    payload_context,
)


class HeaderValidationMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        x_api_key_1: str,
        x_api_key_2: str,
        excluded_paths: FrozenSet[str] = frozenset(),
        authexpiryignore_paths: FrozenSet[str] = frozenset(),
    ):
        self.app = app
        self.x_api_key_1 = x_api_key_1
        self.x_api_key_2 = x_api_key_2
        self.excluded_paths: FrozenSet[str] = frozenset(
            {"/docs", "/openapi.json", *excluded_paths}
        )
        self.authexpiryignore_paths: FrozenSet[str] = frozenset(authexpiryignore_paths)

    async def validate_headers(self, headers, path: str) -> tuple[bool, dict, int]:
        """Helper method to validate headers for both HTTP and WebSocket."""
        access_token = None
        verify_exp = True
        should_ignore_expiry = any(
            path.startswith(excluded_path)
            for excluded_path in self.authexpiryignore_paths
        )

        # Check for access token in headers
        if "access_token" in headers.get("cookie", ""):
            # Parse cookie string to get access token
            cookies = dict(
                cookie.split("=")
                for cookie in headers.get("cookie", "").split("; ")
                if cookie
            )
            access_token = cookies.get("access_token")
            verify_exp = not should_ignore_expiry
        elif "authorization" in headers:
            access_token = headers["authorization"]
            verify_exp = not should_ignore_expiry

        try:
            if access_token:
                payload = JwtValdationUtils.validate_token(
                    access_token, verify_exp=verify_exp
                )
                headers_model = Headers_PM(
                    **dict(
                        correlationid=headers.get("correlationid", str(uuid.uuid4())),
                        username=payload["username"],
                        authorization=access_token,
                    )
                )
                headers_context.set(headers_model)
                return True, {}, 200

            if (
                headers.get("x-api-key-1") == self.x_api_key_1
                or headers.get("x-api-key-2") == self.x_api_key_2
            ):
                headers_model = Headers_PM(
                    **dict(
                        correlationid=headers.get("correlationid", str(uuid.uuid4())),
                        username=headers.get("username", ""),
                        authorization=headers.get("authorization", ""),
                    )
                )
                headers_context.set(headers_model)
                return True, {}, 200

            return (
                False,
                {"detail": "Unauthorized: Missing or invalid credentials."},
                401,
            )

        except HTTPException as e:
            return False, {"detail": str(e.detail)}, e.status_code
        except Exception as e:
            APP_LOGGER.error(f"Unexpected error in token validation: {str(e)}")
            return False, {"detail": "Internal server error"}, 500

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        if scope["type"] not in ["http", "websocket"]:
            return await self.app(scope, receive, send)

        path = scope["path"]
        if any(path.startswith(excluded_path) for excluded_path in self.excluded_paths):
            return await self.app(scope, receive, send)

        headers = dict(scope["headers"])
        headers = {k.decode(): v.decode() for k, v in headers.items()}

        is_valid, error_detail, status_code = await self.validate_headers(headers, path)

        if not is_valid:
            if scope["type"] == "http":
                response = JSONResponse(error_detail, status_code=status_code)
                return await response(scope, receive, send)
            else:

                async def close_websocket():
                    await send(
                        {
                            "type": "websocket.close",
                            "code": 4000
                            + status_code,  # Using 4000 + HTTP status code as WebSocket close code
                            "reason": error_detail["detail"],
                        }
                    )

                return await close_websocket()

        return await self.app(scope, receive, send)


class ExceptionMiddleware:
    def __init__(self, app: ASGIApp):
        self.app = app

    def log_exception(self):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        if exc_type is not None:
            traceback_info = traceback.extract_tb(exc_traceback)
            relevant_traceback = next(
                (
                    trace
                    for trace in reversed(traceback_info)
                    if str(config("SERVICENAME")) in trace.filename
                ),
                None,
            )
            if relevant_traceback:
                pathname, lineno, funcName, code_line = relevant_traceback
                error_data = {
                    "exception_type": str(exc_type.__name__),
                    "exception_value": str(exc_value),
                    "pathname": str(pathname),
                    "lineno": str(lineno),
                    "funcName": funcName,
                    "code_line": str(code_line),
                }
                APP_LOGGER.error(json.dumps(error_data))

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        try:
            await self.app(scope, receive, send)

        except HTTPException as http_exception:
            self.log_exception()
            await self.handle_http_exception(http_exception, scope, receive, send)
        except Exception as e:
            self.log_exception()
            await self.handle_internal_server_error(e, scope, receive, send)

    async def handle_http_exception(self, http_exception, scope, receive, send):
        response = JSONResponse(
            {"detail": http_exception.detail},
            status_code=http_exception.status_code,
        )
        await response(scope, receive, send)

    async def handle_internal_server_error(self, exception, scope, receive, send):
        error_message = {
            "detail": "Internal Server Error",
            "error_message": str(exception),
        }
        response = JSONResponse(status_code=500, content=error_message)
        await response(scope, receive, send)


class ContextSetter:
    async def __call__(self, request: Request):
        if request.query_params:
            query_params = dict(request.query_params)
            request_params_context.set(query_params)
        try:
            body = await request.json()
            payload_context.set(body)
        except (json.JSONDecodeError, ValueError):
            pass

        request_path_method: RequestContext_PM = RequestContext_PM(
            request_trace_id=str(uuid.uuid4()),
            url_path=request.url.path,
            method=request.method,
        )
        request_context.set(request_path_method)



import sys
import traceback
from typing import Any, Dict, Type, cast
from pydantic import BaseModel
from decouple import config

from ..context.vars import headers_context, payload_context
from ..interfaces.interfaces_pd import Headers_PM
from ..interfaces.interfaces_th import Headers_TH
from ..utils.jwt_validation import JwtValdationUtils
from ..logging.base_logger import APP_LOGGER


class MessageMiddleware:
    @staticmethod
    def validate_set_context(
        payload: Dict[str, Any],
        headers: Headers_TH,
        payload_pydantic_model: Type[BaseModel],
    ):
        try:
            if headers.get("authorization"):
                JwtValdationUtils.validate_token(
                    token=cast(str, headers.get("authorization")),
                    verify_aud=False,
                    verify_exp=False,
                )
            pydantic_payload = payload_pydantic_model(**payload)
            payload_context.set(pydantic_payload)
            headers_context.set(Headers_PM(**headers))
        except Exception as e:
            raise ValueError(f"Validation or context setting failed: {e}")


class ExceptionLogger:
    @staticmethod
    def log_exception(instance, correlation_id: str) -> None:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        if exc_type is not None:
            traceback_info = traceback.extract_tb(exc_traceback)
            relevant_traceback = next(
                (
                    trace
                    for trace in reversed(traceback_info)
                    if str(config("SERVICENAME")) in trace.filename
                ),
                None,
            )
            error_data = {
                "exception_type": str(exc_type.__name__),
                "exception_value": str(exc_value),
                "correlationid": correlation_id,
            }
            if relevant_traceback:
                pathname, lineno, func_name, code_line = relevant_traceback
                error_data.update(
                    {
                        "pathname": pathname,
                        "lineno": lineno,
                        "funcName": func_name,
                        "code_line": code_line,
                    }
                )
            APP_LOGGER.error(error_data)



from typing import Dict
from ...interfaces.interfaces_th import (
    LLM_ModelConfig_TH,
    BedrockPayload_TH,
    LLM_SpecialTokens_TH,
)

FMS_BEDROCK: Dict[str, LLM_ModelConfig_TH] = {
    "meta.llama3-8b-instruct-v1:0": LLM_ModelConfig_TH(
        modelName="Llama 3 8B Instruct",
        providerName="Meta",
        responseStreamingSupported=True,
        payload_format=BedrockPayload_TH(
            prompt="", max_gen_len=64, temperature=0.6, top_p=0.6
        ),
        special_tokens=LLM_SpecialTokens_TH(
            bos_token="<|begin_of_text|>",
            content_start_token="<|start_header_id|>{role}<|end_header_id|>",
            content_end_token="<|eot_id|>",
            eos_token="<|end_of_text|>",
        ),
    ),
    "meta.llama3-70b-instruct-v1:0": LLM_ModelConfig_TH(
        modelName="Llama 3 70B Instruct",
        providerName="Meta",
        responseStreamingSupported=False,
        payload_format=BedrockPayload_TH(
            prompt="", max_gen_len=64, temperature=0.6, top_p=0.6
        ),
        special_tokens=LLM_SpecialTokens_TH(
            bos_token="<|begin_of_text|>",
            content_start_token="<|start_header_id|>{role}<|end_header_id|>",
            content_end_token="<|eot_id|>",
            eos_token="<|end_of_text|>",
        ),
    ),
    "meta.llama3-1-8b-instruct-v1:0": LLM_ModelConfig_TH(
        modelName="Llama 3 8B Instruct",
        providerName="Meta",
        responseStreamingSupported=True,
        payload_format=BedrockPayload_TH(
            prompt="", max_gen_len=64, temperature=0.6, top_p=0.6
        ),
        special_tokens=LLM_SpecialTokens_TH(
            bos_token="<|begin_of_text|>",
            content_start_token="<|start_header_id|>{role}<|end_header_id|>",
            content_end_token="<|eot_id|>",
            eos_token="<|end_of_text|>",
        ),
    ),
    "meta.llama3-1-70b-instruct-v1:0": LLM_ModelConfig_TH(
        modelName="Llama 3 70B Instruct",
        providerName="Meta",
        responseStreamingSupported=False,
        payload_format=BedrockPayload_TH(
            prompt="", max_gen_len=64, temperature=0.6, top_p=0.6
        ),
        special_tokens=LLM_SpecialTokens_TH(
            bos_token="<|begin_of_text|>",
            content_start_token="<|start_header_id|>{role}<|end_header_id|>",
            content_end_token="<|eot_id|>",
            eos_token="<|end_of_text|>",
        ),
    ),
    "mistral.mistral-7b-instruct-v0:2": LLM_ModelConfig_TH(
        modelName="Mistral 7B Instruct",
        providerName="Mistral AI",
        responseStreamingSupported=True,
        payload_format=BedrockPayload_TH(
            prompt="",
            max_tokens=64,
            temperature=0.6,
            top_p=0.6,
            top_k=6,
            # stop=["."],
        ),
        special_tokens=LLM_SpecialTokens_TH(
            bos_token="<s>",
            content_start_token="[INST]",
            content_end_token="[/INST]",
            eos_token="</s>",
        ),
    ),
    "mistral.mixtral-8x7b-instruct-v0:1": LLM_ModelConfig_TH(
        modelName="Mixtral 8x7B Instruct",
        providerName="Mistral AI",
        responseStreamingSupported=False,
        payload_format=BedrockPayload_TH(
            prompt="",
            max_tokens=64,
            temperature=0.6,
            top_p=0.6,
            top_k=6,
            # stop=["."],
        ),
        special_tokens=LLM_SpecialTokens_TH(
            bos_token="<s>",
            content_start_token="[INST]",
            content_end_token="[/INST]",
            eos_token="</s>",
        ),
    ),
}



# import time
# from functools import wraps
# from typing import Optional

# from fastapi import APIRouter, Depends, HTTPException, Request
# from fastapi.security import HTTPBearer
# from pydantic import BaseSettings
# from redis import Redis
# from redis.exceptions import RedisError

# # Configuration Management
# class Settings(BaseSettings):
#     redis_host: str = "localhost"
#     redis_port: int = 6379
#     default_rate_limit: int = 1000

#     class Config:
#         env_file = ".env"

# settings = Settings()

# # Rate Limiter Class
# class RateLimiter:
#     def __init__(self, redis_client: Redis, default_limit: int = settings.default_rate_limit):
#         self.redis_client = redis_client
#         self.default_limit = default_limit

#     def check_rate_limit(self, key: str, max_calls: Optional[int] = None):
#         """Checks if the request exceeds the daily rate limit."""
#         max_calls = max_calls or self.default_limit
#         key_with_date = f"{key}:{time.strftime('%Y-%m-%d')}"  # Key for today's count

#         try:
#             # Increment the call count
#             current_count = self.redis_client.incr(key_with_date)

#             # Set expiration if it's the first call
#             if current_count == 1:
#                 self.redis_client.expire(key_with_date, 86400)  # 86400 seconds = 24 hours

#             if current_count > max_calls:
#                 raise HTTPException(
#                     status_code=429,
#                     detail=f"Rate limit exceeded. Maximum {max_calls} requests per day allowed for this endpoint."
#                 )
#         except RedisError as e:
#             raise HTTPException(
#                 status_code=500,
#                 detail="Internal server error while processing rate limit."
#             )

#     async def rate_limit_middleware(self, request: Request, max_calls: Optional[int] = None):
#         """Apply rate limiting dynamically based on request path and method."""
#         key = f"{request.method}:{request.url.path}"
#         self.check_rate_limit(key, max_calls)

# # Dependency Injection for Rate Limiter
# def get_redis_client():
#     return Redis(host=settings.redis_host, port=settings.redis_port, decode_responses=True)

# def get_rate_limiter(redis_client: Redis = Depends(get_redis_client)):
#     return RateLimiter(redis_client)

# # Rate Limited Decorator
# def rate_limited(limiter: RateLimiter, max_calls: Optional[int] = None):
#     def decorator(func):
#         @wraps(func)
#         async def wrapper(request: Request, *args, **kwargs):
#             await limiter.rate_limit_middleware(request, max_calls)
#             return await func(request, *args, **kwargs)
#         return wrapper
#     return decorator

# # Router and Endpoints
# router = APIRouter(
#     prefix="/api/v1/authorization",
#     tags=["Auth"],
# )

# @router.post("/signup")
# @rate_limited(limiter=get_rate_limiter(), max_calls=500)
# async def signup(request: Request, payload: SignUp_PM):
#     return AuthController.signup(payload=payload)

# @router.post("/login")
# @rate_limited(limiter=get_rate_limiter(), max_calls=300)
# async def login(request: Request, payload: Login_PM):
#     return AuthController.login(payload=payload)



from .bedrock_client import GenerateBedrockResponse
from .local_interfaces_typehinter import LLM_ClientPayload_TH
from ...interfaces.interfaces_th import (
    LLM_HyperParameters_TH,
    LLM_PromptTemplates_TH,
)
from .base_llm_operation import LLMOperation, BaseAIEventsConfig
from .foundation_models import FoundationModels



import json
import datetime
import logging
import pytz
from decouple import config

# modules
from ..context.vars import (
    headers_context,
    request_context,
    payload_context,
    request_params_context,
)
from typing import Any, Dict, Optional


class ContextManager:

    @staticmethod
    def _get_model_dump(obj):
        if hasattr(obj, "model_dump"):
            return obj.model_dump(mode="json")
        elif isinstance(obj, dict):
            return obj
        return None

    @classmethod
    def get_current_context(cls) -> Dict[str, Any]:
        """
        Retrieve the current request-specific contexts
        """
        return {
            "headers_context": cls._get_model_dump(headers_context.get()),
            "payload_context": cls._get_model_dump(payload_context.get()),
            "request_context": cls._get_model_dump(request_context.get()),
            "request_params_context": cls._get_model_dump(request_params_context.get()),
        }


class RecordFormatter:
    DATE_FORMAT_TIMEZONE = "%a %b %d %Y %H:%M:%S (%Z)"

    @staticmethod
    def format_record(record: logging.LogRecord) -> str:
        record_items = dict(record.__dict__)
        try:
            message_dict = json.loads(str(record_items.get("message")))
            record_items.update(message_dict)
            record_items["message"] = json.dumps(
                {
                    key: value
                    for key, value in message_dict.items()
                    if key not in ["pathname", "lineno", "funcName"]
                }
            )
        except (ValueError, TypeError):
            pass

        output = {
            "levelname": record_items["levelname"],
            "epoch": record_items["created"],
            "timestamp": datetime.datetime.fromtimestamp(
                record_items["created"], tz=pytz.timezone("Asia/Kolkata")
            ).strftime(RecordFormatter.DATE_FORMAT_TIMEZONE),
            "message": record_items["message"],
            "serviceName": config("SERVICENAME"),
            "env": config("ENVIRONMENT"),
            "funcName": record_items["funcName"],
            "pathname": record_items["pathname"],
            "lineno": record_items["lineno"],
        }
        context = ContextManager.get_current_context()
        output.update(context)
        return json.dumps(output)


class CustomJsonFormatter(logging.Formatter):

    def format(self, record: logging.LogRecord) -> str:
        super().format(record)
        return RecordFormatter.format_record(record)


class LoggerFactory:
    _logger: Optional[logging.Logger] = None

    @classmethod
    def get_logger(cls) -> logging.Logger:
        """
        Get or create a logger instance
        """
        if cls._logger is None:
            cls._logger = logging.getLogger("fastapi_app")
            cls._logger.setLevel(logging.DEBUG)
            cls._logger.handlers.clear()

            # File handler
            file_handler = logging.FileHandler(str(config("LOGGING_FILENAME")))
            file_handler.setFormatter(CustomJsonFormatter())
            cls._logger.addHandler(file_handler)
            console_handler = logging.StreamHandler()
            console_handler.setFormatter(CustomJsonFormatter())
            cls._logger.addHandler(console_handler)

        return cls._logger


APP_LOGGER = LoggerFactory.get_logger()



from typing import TypedDict, Required, Dict

from typing import TypedDict, List, Dict, Required


class Jwk_TH(TypedDict, total=False):
    alg: Required[str]
    e: str
    kid: Required[str]
    kty: str
    n: str
    use: str


class Headers_TH(TypedDict):
    correlationid: str
    username: str
    authorization: str


class SQSClientCallBackResponse_TH(TypedDict):
    allSuccess: bool
    correlationid: str


class LLM_HyperParameters_TH(TypedDict, total=False):
    max_gen_len: int
    temperature: Required[float]
    top_p: Required[float]
    top_k: int
    max_tokens: int
    stop: List[str]


class BedrockPayload_TH(TypedDict, total=False):
    prompt: str
    max_gen_len: int
    temperature: Required[float]
    top_p: Required[float]
    top_k: int
    max_tokens: int
    stop: List[str]


class LLM_SpecialTokens_TH(TypedDict):
    bos_token: str
    content_start_token: str
    content_end_token: str
    eos_token: str


class LLM_ModelConfig_TH(TypedDict):
    modelName: str
    providerName: str
    responseStreamingSupported: bool
    payload_format: BedrockPayload_TH
    special_tokens: LLM_SpecialTokens_TH


class LLM_PromptTemplates_TH(TypedDict, total=False):
    system_prompt: str
    user_prompt: Required[str]
    user_prompt_1: str
    base_token_count: int



import boto3
import json
from typing import Dict, Union
from decouple import config


class BedrockClient_Sync:

    def __init__(self) -> None:
        self.client = boto3.client(
            "bedrock-runtime",
            region_name=str(config("AWS_REGION_NAME")),
            aws_access_key_id=str(config("AWS_ACCESS_KEY")),
            aws_secret_access_key=str(config("AWS_SECRET_ACCESS_KEY")),
        )

    def no_stream_bedrock_response(
        self, payload: dict, modelId: str
    ) -> Dict[str, Union[str, int]]:
        try:
            response = self.client.invoke_model(
                body=json.dumps(payload),
                modelId=modelId,
                accept="application/json",
                contentType="application/json",
            )
            body = response.get("body").read().decode("utf-8")
            response_body = json.loads(body)
            result = (
                response_body.get("generation")
                or response_body.get("outputs")[0]["text"]
            )
            tokens_count = response_body.get("generation_token_count") or None
            if tokens_count:
                return dict(llm_response=result, tokens=tokens_count)
            return dict(llm_response=result)
        except (
            self.client.exceptions.AccessDeniedException,
            self.client.exceptions.ResourceNotFoundException,
            self.client.exceptions.ThrottlingException,
            self.client.exceptions.ModelTimeoutException,
            self.client.exceptions.InternalServerException,
            self.client.exceptions.ModelStreamErrorException,
            self.client.exceptions.ValidationException,
            self.client.exceptions.ModelNotReadyException,
            self.client.exceptions.ServiceQuotaExceededException,
            self.client.exceptions.ModelErrorException,
        ) as e:
            raise e

    def stream_bedrock_response(self, payload=None, modelId=None):
        try:
            response_stream = self.client.invoke_model_with_response_stream(
                body=json.dumps(payload),
                modelId=modelId,
                accept="application/json",
                contentType="application/json",
            )
            return response_stream
        except (
            self.client.exceptions.AccessDeniedException,
            self.client.exceptions.ResourceNotFoundException,
            self.client.exceptions.ThrottlingException,
            self.client.exceptions.ModelTimeoutException,
            self.client.exceptions.InternalServerException,
            self.client.exceptions.ModelStreamErrorException,
            self.client.exceptions.ValidationException,
            self.client.exceptions.ModelNotReadyException,
            self.client.exceptions.ServiceQuotaExceededException,
            self.client.exceptions.ModelErrorException,
        ) as e:
            raise e



from typing import Dict, Tuple
from decouple import UndefinedValueError, config


class Helpers:
    _url_cache: Dict[Tuple[str, str, str, str], str] = {}

    @staticmethod
    def construct_queue_url(
        queue_name: str, region_name: str, params: Dict[str, bool], aws_account_id: str
    ) -> str:
        # Create a unique key for caching
        is_fifo = params.get("fifo", False)
        cache_key = (queue_name, region_name, aws_account_id, str(is_fifo))

        # Check if URL is already cached
        if cache_key in Helpers._url_cache:
            return Helpers._url_cache[cache_key]

        # Fetch environment
        try:
            __environment: str = str(config("ENVIRONMENT"))
        except UndefinedValueError:
            raise ValueError(
                "Specify the name of the environment within the '.env' file."
            )

        # Determine queue name
        try:
            __queue_name: str = str(config(queue_name))
        except UndefinedValueError:
            __queue_name = queue_name
            if not __queue_name:
                raise ValueError(
                    f"The following queue name does not exist: {queue_name}"
                )

        # Construct the queue URL
        queue_url = (
            "https://sqs.{AWS_REGION_NAME}.amazonaws.com/{AWS_ACCOUNT_ID}/queue-{AWS_REGION_NAME}-{QUEUE_NAME}-{ENVIRONMENT}"
        ).format(
            AWS_REGION_NAME=region_name,
            AWS_ACCOUNT_ID=aws_account_id,
            QUEUE_NAME=__queue_name,
            ENVIRONMENT=__environment,
        )

        # Append .fifo if needed
        if is_fifo:
            queue_url += ".fifo"

        # Cache the constructed URL
        Helpers._url_cache[cache_key] = queue_url

        return queue_url



from setuptools import setup, find_packages

# Read requirements from the file
with open("requirements.txt") as f:
    requirements = f.read().splitlines()

# Read the README for long description
with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="zoldics_service_utils",
    version="0.0.5",
    author="Nabarka_Deb",
    description="A most utile package for the governance of pythonic micro-services",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/BRIZINGR007/PyhtonServiceUtil",
    package_dir={"": "app"},
    packages=find_packages(where="app"),
    install_requires=requirements,
    extra_require={"dev": ["pytest>=7.0", "twine>=4.0.2"]},
    python_requires=">=3.11",
)



import asyncio
from redis.asyncio import Redis
import json
from typing import AsyncGenerator, Optional, Any, List, cast

from ...logging.base_logger import APP_LOGGER

from ...interfaces.interfaces_pd import SSEPayload_PM
from ...ioc.singleton import SingletonMeta


class AsyncRedisClient(metaclass=SingletonMeta):
    def __init__(self, redis_connection: Optional[Redis] = None) -> None:
        if not hasattr(self, "redis") and isinstance(redis_connection, Redis):
            self.redis: Redis = redis_connection
        self.pubsub = self.redis.pubsub()
        self._lock = asyncio.Lock()  # Add lock for thread-safe operations

    async def set(self, key: str, value: Any, expiry: Optional[int] = None) -> bool:
        return bool(await self.redis.set(key, value, ex=expiry))

    async def get(self, key: str) -> Optional[bytes]:
        return await self.redis.get(key)

    async def delete(self, key: str) -> int:
        return await self.redis.delete(key)

    async def delete_all_keys(self, keys: List[str]) -> int:
        return await self.redis.delete(*keys)

    async def get_keys(self, pattern: str = "*") -> List[str]:
        return cast(List[str], await self.redis.keys(pattern))

    async def send_command(self, *commands: str) -> Any:
        return await self.redis.execute_command(*commands)

    async def publish(self, channel: str, message: SSEPayload_PM) -> int:
        return await self.redis.publish(
            channel=channel, message=json.dumps(message.model_dump())
        )

    async def subscribe(self, channel: str) -> None:
        async with self._lock:
            await self.pubsub.subscribe(channel)

    async def unsubscribe(self, channel: str) -> None:
        async with self._lock:
            await self.pubsub.unsubscribe(channel)

    async def get_message(self, timeout: float = 1.0) -> Optional[dict]:
        async with self._lock:
            return await self.pubsub.get_message(timeout=timeout)

    async def listen(self, channel: str) -> AsyncGenerator[dict, None]:
        await self.subscribe(channel=channel)
        try:
            while True:
                try:
                    async with self._lock:
                        message = await self.pubsub.get_message(timeout=1.0)
                        if message is not None:
                            yield message
                    await asyncio.sleep(0.1)
                except Exception as e:
                    await self.unsubscribe(channel)
                    APP_LOGGER.error(f"Error in listen loop: {e}")
                    await asyncio.sleep(1)
        finally:
            await self.unsubscribe(channel)

    async def close_connection(self) -> None:
        async with self._lock:
            await self.pubsub.close()
            await self.redis.close()



import json
from typing import Dict, List, Optional, Union
from fastapi.responses import StreamingResponse

from ..bedrock_client.local_interfaces_typehinter import (
    LLM_ClientPayload_TH,
)

from ..bedrock_client.call_bedrock import BedrockClient_Sync
from ..bedrock_client.fm_config import FMS_BEDROCK
from ...interfaces.interfaces_th import (
    LLM_HyperParameters_TH,
    LLM_SpecialTokens_TH,
)
from ...logging.base_logger import APP_LOGGER


class GenerateBedrockResponse:

    def __init__(
        self,
        payload: LLM_ClientPayload_TH,
        streaming: bool,
    ) -> None:
        self.bedrock_client = BedrockClient_Sync()
        self.streaming: bool = streaming
        self.messages: List[Dict[str, str]] = payload["messages"]
        self.hyper_parameters: LLM_HyperParameters_TH = payload["model_hyperparameters"]
        self.model_id = payload["modelId"].value

    def __construct_prompt_template(self, modelId: str, messages: List[dict]):
        special_tokens: LLM_SpecialTokens_TH = FMS_BEDROCK.get(modelId, {}).get(
            "special_tokens", {}
        )
        bos_token: str = str(special_tokens.get("bos_token"))
        content_start_token: str = str(special_tokens.get("content_start_token"))
        content_end_token: str = str(special_tokens.get("content_end_token"))
        provider_name = FMS_BEDROCK.get(modelId, {}).get("providerName")

        def get_start_token(role):
            if provider_name == "Meta":
                return content_start_token.format(role=role)
            elif provider_name == "Mistral AI":
                return content_start_token
            return ""

        structured_prompt = f"{bos_token}" + "\n"

        for msg in messages:
            role: str = str(msg.get("role"))
            content: str = str(msg.get("content"))
            if role in {"system", "user"}:
                structured_prompt += get_start_token(role) + "\n"
                structured_prompt += content + "\n" + content_end_token + "\n"
        if provider_name == "Meta":
            structured_prompt += "<|start_header_id|>assistant<|end_header_id|>"
        return structured_prompt

    def payload_streaming(self, provider_name: str, response_stream: dict):
        if provider_name == "Meta":
            for event in response_stream.get("body", []):
                response = json.loads(event["chunk"]["bytes"]).get("generation")
                yield json.dumps({"chunk": response}) + "\n"

        elif provider_name == "Mistral AI":
            for event in response_stream.get("body", []):
                response = (
                    json.loads(event["chunk"]["bytes"]).get("outputs")[0].get("text")
                )
                yield json.dumps({"chunk": response}) + "\n"

    def execute(self) -> Optional[Union[StreamingResponse, Dict[str, Union[str, int]]]]:
        try:
            constructed_prompt: str = self.__construct_prompt_template(
                modelId=self.model_id, messages=self.messages
            )
            payload = dict(prompt=constructed_prompt, **self.hyper_parameters)
            if self.streaming:
                provider_name: str = str(
                    FMS_BEDROCK.get(self.model_id, {}).get("providerName")
                )
                response_stream = self.bedrock_client.stream_bedrock_response(
                    payload=payload, modelId=self.model_id
                )
                return StreamingResponse(
                    self.payload_streaming(
                        provider_name, response_stream=response_stream
                    ),
                    media_type="text/event-stream",
                )
            else:
                llm_response: Dict[str, Union[str, int]] = (
                    self.bedrock_client.no_stream_bedrock_response(
                        payload=payload, modelId=self.model_id
                    )
                )
                return llm_response

        except Exception as exception:
            APP_LOGGER.error(str(exception))
        return {}



from typing import cast

from ..interfaces.interfaces_pd import Headers_PM
from ..interfaces.interfaces_th import Headers_TH
from ..context.vars import headers_context


class ContextUtils:
    @staticmethod
    def get_header_details() -> Headers_PM:
        return cast(Headers_PM, headers_context.get())



from decouple import config
from mongoengine import get_connection
from pymongo import MongoClient
from pymongo.client_session import ClientSession
import contextvars
from typing import Optional, Any, Type, TypeVar, cast
from contextlib import suppress

# Global context variable for MongoDB session
mongo_session_context: contextvars.ContextVar[Optional[ClientSession]] = (
    contextvars.ContextVar("mongo_session", default=None)
)


class MongoTransactionContext:
    """Context manager for handling MongoDB transactions with proper session management."""

    def __init__(self, db_alias: Optional[str] = None):
        """Initialize the transaction context.

        Args:
            db_alias: Optional database alias. If not provided, fetches from config.
        """
        self.db_alias = db_alias or cast(str, config("DB_NAME"))
        self.session: Optional[ClientSession] = None
        self.token: Optional[contextvars.Token[Optional[ClientSession]]] = None

    @property
    def client(self) -> MongoClient:
        """Get the MongoDB client instance."""
        connection = get_connection(alias=self.db_alias)

        if not hasattr(connection, "connection") or not hasattr(
            connection.connection, "client"
        ):
            raise ValueError(f"Invalid MongoDB connection for alias {self.db_alias}")

        return connection.connection.client

    def __enter__(self) -> ClientSession:
        """Enter the transaction context.

        Returns:
            Active MongoDB client session.

        Raises:
            ValueError: If MongoDB connection is invalid.
            RuntimeError: If session creation fails.
        """
        try:
            self.session = self.client.start_session()
            self.session.start_transaction()
            self.token = mongo_session_context.set(self.session)
            return self.session

        except Exception as e:
            # Clean up any partially initialized resources
            if self.session:
                with suppress(Exception):
                    self.session.end_session()
            raise RuntimeError(f"Failed to create MongoDB session: {str(e)}") from e

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[Any],
    ) -> None:
        if not self.session:
            return

        try:
            if exc_type is None:
                self.session.commit_transaction()
            else:
                self.session.abort_transaction()

        except Exception:
            with suppress(Exception):
                self.session.abort_transaction()
            raise

        finally:
            if self.token is not None:
                mongo_session_context.reset(self.token)

            with suppress(Exception):
                self.session.end_session()

            self.session = None
            self.token = None
